# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
A script that attempts to reveal information about the spectral uniformity accross the field of view of the microscope.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pwspy.dataTypes import ExtraReflectanceCube, ExtraReflectionCube, ImCube
from pwspy.utility.reflection import reflectanceHelper, Material
from glob import glob
import os
from pwspy.utility import thinFilmPath
from pwspy.dataTypes import Roi

#TODO change the theoretical CSV file to use BK7 instead of SIO2. Our measurements are still not matching theory though. Is this because of all the extra scattering?

__spec__ = None


def colorbar(mappable):
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    ax = mappable.axes
    fig = ax.figure
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    return fig.colorbar(mappable, cax=cax)

if __name__ == "__main__":
    # %%User Input
    wDir = r'J:\Calibrations\ITOThinFilm\LCPWS2'
    referenceMaterial = Material.Water
    referenceNumber = 999
    cellNumber = 1
    
    files = glob(os.path.join(wDir, '*'))
    files = [f for f in files if os.path.isdir(f)]

    plt.ion()

    # %% Loading
    theory = pd.read_csv(os.path.join(thinFilmPath, '2um_sio2_ITO_air.txt'), delimiter='\t', index_col=0)
    theory = theory.loc[500:700:2]
    theory = theory
    ztheory = (theory - theory.mean()) / (theory.std() * theory.shape[0])  # Zero normalize for correlation
    sref = reflectanceHelper.getReflectance(referenceMaterial, Material.Glass, index=range(500, 701, 2))

    result = []
    for f in files:
        print("Loading")
        im = ImCube.loadAny(os.path.join(f, f'Cell{cellNumber}'))
        ref = ImCube.loadAny(os.path.join(f, f'Cell{referenceNumber}'))
        er = ExtraReflectanceCube.fromHdfFile(r'C:\Users\backman05\PwspyApps\ExtraReflectanceCreatorData\GoogleDriveData', 'LCPWS2_100xpfs-7_8_2019')
        print("Dividing data by mirror spectra")
        im.correctCameraEffects()
        ref.correctCameraEffects()
        im.normalizeByExposure()
        ref.normalizeByExposure()
        Iextra = ExtraReflectionCube.create(er, sref, ref)
        im.subtractExtraReflection(Iextra)
        ref.subtractExtraReflection(Iextra)
        mSpectra = ref.getMeanSpectra()[0] # we just want an average spectra to normalize by so we don't normalize out effects that we want to see.
        im /= mSpectra[None, None, :]
        im *= np.array(sref) # Normalize to physical reflectance
        zim = (im.data - im.data.mean(axis=2, keepdims=True)) / (im.data.std(axis=2, keepdims=True))  # zero normalize for correlation
        print("Cross correlating with theory")
        corr = np.apply_along_axis(lambda arr: np.correlate(np.array(ztheory).squeeze(), arr, mode='same'), axis=2, arr=zim)

        # Data Extraction
        print("Extracting data from correlation")
        amp = im.data.max(axis=2) - im.data.min(axis=2)  # The amplitude of the spectra.
        corramp = np.max(corr, axis=2)  # The strength of the correlation
        corrmax = np.argmax(corr, axis=2) - corr.shape[2] // 2  # The shift in index that gives the strongest correlation.
        minn = (im.data.min(axis=2) - np.array(
            theory).min())  # The difference between the minimum of the spectra and the theoretical minima. this indicates unwanted light.
        result.append((im, amp, corramp, corrmax, minn))


    for im, amp, corramp, corrmax, minn in result:
        # Plotting
        fig, axs = plt.subplots(nrows=2, ncols=2)
        fig.suptitle(os.path.split(im.metadata.filePath)[-1])
        imsh = axs[0, 0].imshow(amp, vmin=np.percentile(amp, 1), vmax=np.percentile(amp, 99))
        colorbar(imsh)
        axs[0, 0].set_title("Amplitude")
        imsh = axs[0, 1].imshow(corrmax, vmin=np.percentile(corrmax, 1), vmax=np.percentile(corrmax, 99))
        colorbar(imsh)
        axs[0, 1].set_title("Shift (index) to max correlation")
        imsh = axs[1, 1].imshow(corramp, vmin=0, vmax=1)
        colorbar(imsh)
        axs[1, 1].set_title("Max Correlation")
        imsh = axs[1, 0].imshow(minn, vmin=np.percentile(minn, 1), vmax=np.percentile(minn, 99))
        colorbar(imsh)
        axs[1, 0].set_title("Minimum of data")
        for ax in axs.flatten():
            ax.axis('off')

    while True:
        try:
            mask = Roi.fromVerts(result[0][0].selectLassoROI(), dataShape=result[0][0].data.shape[:-1])
            fig2, ax = plt.subplots()
            ax.plot(np.array(theory.index).astype(np.float), np.array(theory["Reflectance"]).astype(np.float),
                    label="Theory")
            for im, amp, corramp, corrmax, minn in result:
                mean, std = im.getMeanSpectra(mask)
                ax.plot(im.wavelengths, mean, label=os.path.split(im.metadata.filePath)[-1])
                ax.fill_between(im.wavelengths, mean - std, mean + std, alpha=0.4)
            ax.legend()
            ax.set_xlabel("Wavelength (nm)")
            ax.set_ylabel("Reflectance")
            fig2.suptitle("Thin Film Comparison")
            fig2.show()
            plt.pause(0.1)
            while plt.fignum_exists(fig2.number):
                fig2.canvas.flush_events()
        except Exception as e:
            raise e
    



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
This script is based on a matlab script written by Lusik Cherkezyan for Nanocytomics.
Nano uses this method to extract rms from phantom make from ChromEM cells embedded in resin.
The phantom has a strong thin-film spectrum. This script is meant to filter out the thin film components
of the fourier transfrom and extract RMS from what is left.
"""

# Last tested on pwspy ad745ef0c1ae526e25981b0036f460981397a456


if __name__ == '__main__':
    from pwspy.dataTypes import CameraCorrection, AcqDir, Roi, ImCube, KCube
    import matplotlib.pyplot as plt
    import scipy.signal as sps
    import os
    import numpy as np

    '''User Input'''
    path = r'2_7_2019 11.07'
    refName = 'Cell999'  # This is an imcube of glass, used for normalization.
    cellNames = ['Cell1', 'Cell2']  # , 'Cell3', 'Cell4','Cell5']
    maskSuffix = 'resin'

    # identify the depth in um to which the OPD spectra need to be integrated
    integrationDepth = 2.0  ##  in um
    isHannWindow = True  # Should Hann windowing be applied to eliminate edge artifacts?
    subtractResinOpd = True
    resetResinMasks = False
    wvStart = 510  # start wavelength for poly subtraction
    wvEnd = 690  # end wavelength for poly subtraction
    sampleRI = 1.545  # The refractive index of the resin. This is taken from matlab code, I don't know if it's correct.
    orderPolyFit = 0
    wv_step = 2
    correction = CameraCorrection(2000, (0.977241216, 1.73E-06, 1.70E-11))

    '''************'''

    b, a = sps.butter(6, 0.1 * wv_step)
    opdIntegralEnd = integrationDepth * 2 * sampleRI  # We need to convert from our desired depth into an opd value. There are some questions about having a 2 here but that's how it is in the matlab code so I'm keeping it.

    ### load and save mirror or glass image cube
    ref = ImCube.fromMetadata(AcqDir(os.path.join(path, refName)).pws)
    ref.correctCameraEffects(correction)
    ref.filterDust(6, pixelSize=1)
    ref.normalizeByExposure()

    if subtractResinOpd:
        ### load and save reference empty resin image cube
        fig, ax = plt.subplots()
        resinOpds = {}
        for cellName in cellNames:
            resin = ImCube.fromMetadata(AcqDir(os.path.join(path, cellName)).pws)
            resin.correctCameraEffects(correction)
            resin.normalizeByExposure()
            resin /= ref
            resin = KCube.fromImCube(resin)
            if resetResinMasks:
                [resin.metadata.acquisitionDirectory.deleteRoi(name, num) for name, num, fformat in resin.metadata.acquisitionDirectory.getRois() if name == maskSuffix]
            if maskSuffix in [name for name, number, fformat in resin.metadata.acquisitionDirectory.getRois()]:
                resinRoi = resin.metadata.acquisitionDirectory.loadRoi(maskSuffix, 1)
            else:
                print('Select a region containing only resin.')
                resinRoi = Roi.fromVerts(resin.selectLassoROI(), resin.data.shape[:2])
                resin.metadata.acquisitionDirectory.saveRoi(maskSuffix, 1, resinRoi)
            resin.data -= resin.data.mean(axis=2)[:, :, np.newaxis]
            opdResin, xvals = resin.getOpd(isHannWindow, indexOpdStop=None, mask=resinRoi.mask)
            resinOpds[cellName] = opdResin
            ax.plot(xvals, opdResin, label=cellName)
            ax.vlines([opdIntegralEnd], ymin=opdResin.min(), ymax=opdResin.max())
            ax.set_xlabel('OPD')
            ax.set_ylabel("Amplitude")
        ax.legend()
        plt.pause(0.2)

    rmses = {}  # Store the rms maps for later saving
    for cellName in cellNames:
        cube = ImCube.fromMetadata(AcqDir(os.path.join(path, cellName)).pws)
        cube.correctCameraEffects(correction)
        cube.normalizeByExposure()
        cube /= ref
        cube.data = sps.filtfilt(b, a, cube.data, axis=2)
        cube = KCube.fromImCube(cube)

        ## -- Polynomial Fit
        print("Subtracting Polynomial")
        polydata = cube.data.reshape((cube.data.shape[0] * cube.data.shape[1], cube.data.shape[2]))
        polydata = np.rollaxis(polydata, 1)  # Flatten the array to 2d and put the wavenumber axis first.
        cubePoly = np.zeros(polydata.shape)  # make an empty array to hold the fit values.
        polydata = np.polyfit(cube.wavenumbers, polydata,
                              orderPolyFit)  # At this point polydata goes from holding the cube data to holding the polynomial values for each pixel. still 2d.
        for i in range(orderPolyFit + 1):
            cubePoly += (np.array(cube.wavenumbers)[:, np.newaxis] ** i) * polydata[i,
                                                                           :]  # Populate cubePoly with the fit values.
        cubePoly = np.moveaxis(cubePoly, 0, 1)
        cubePoly = cubePoly.reshape(cube.data.shape)  # reshape back to a cube.
        # Remove the polynomial fit from filtered cubeCell.
        cube.data = cube.data - cubePoly

        rmsData = np.sqrt(np.mean(cube.data ** 2, axis=2)) #This can be compared to rmsOPDIntData, when the integralStopIdx is high and we don't do spectral subtraction they should be equivalent.

        # Find the fft for each signal in the desired wavelength range
        opdData, xvals = cube.getOpd(isHannWindow, None)

        if subtractResinOpd:
            opdData = opdData - resinOpds[cellName]

        try:
            integralStopIdx = np.where(xvals >= opdIntegralEnd)[0][0]
        except IndexError:  # If we get an index error here then our opdIntegralEnd is probably bigger than we can achieve. Just use the biggest value we have.
            integralStopIdx = None
            opdIntegralEnd = max(xvals)
            print(f'Integrating to OPD {opdIntegralEnd}')

        opdSquared = np.sum(opdData[:, :, :integralStopIdx] ** 2,axis=2)  # Parseval's theorem tells us that this is equivalent to the sum of the squares of our original signal
        opdSquared *= len(cube.wavenumbers) / opdData.shape[2]  # If the original data and opd were of the same length then the above line would be correct. Since the fft has been upsampled. we need to normalize.
        rmsOpdIntData = np.sqrt(opdSquared)  # this should be equivalent to normal RMS if our stop index is high and resin subtraction is disabled.

        cmap = plt.get_cmap('jet')
        fig, axs = plt.subplots(1, 2, sharex=True, sharey=True)
        im = axs[0].imshow(rmsData, cmap=cmap, clim=[np.percentile(rmsData, 0.5), np.percentile(rmsData, 99.5)])
        fig.colorbar(im, ax=axs[0])
        axs[0].set_title('RMS')
        im = axs[1].imshow(rmsOpdIntData, cmap=cmap,
                           clim=[np.percentile(rmsOpdIntData, 0.5), np.percentile(rmsOpdIntData, 99.5)])
        fig.colorbar(im, ax=axs[1])
        axs[1].set_title(f'RMS from OPD below {opdIntegralEnd} after resin OPD subtraction')
        fig.suptitle(cellName)
        rmses[cellName] = rmsOpdIntData
        plt.pause(0.2)
    plt.pause(0.5)

    ## plt.waitforbuttonpress(timeout=-1)
    #for k, v in rmses.items():
    #    if input(f"Save opdRms for {k}? (y/n): ").strip().lower() == 'y':
    #        sp.io.savemat(os.path.join(path, k, 'phantom_Rms.mat'), {'cubeRms': v.astype(np.float32)})  # save as a single



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""A collection of functions dedicated to the purpose of generating Extra Reflectance calibrations from images of
materials with known reflectances (e.g. air/glass interface, water/glass interface.)
These functions are relied on heavily in the pwspy.apps.ERCreator app.

Functions
----------
.. autosummary::
   :toctree: generated/

   getTheoreticalReflectances
   generateMaterialCombos
   getAllCubeCombos
   plotExtraReflection
   generateRExtraCubes

Classes
--------
.. autosummary
   :toctree: generated/

   CubeCombo
"""
import logging
from typing import Dict, List, Tuple, Iterable, Any, Iterator, Union, Optional, Set

from mpl_qt_viz.visualizers import DockablePlotWindow

from pwspy.dataTypes import Roi, ExtraReflectanceCube, ImCube, ERMetaData
from pwspy.utility.reflection import reflectanceHelper, Material
import itertools
import matplotlib.pyplot as plt
import numpy as np
from functools import reduce
import pandas as pd
from dataclasses import dataclass
__all__ = ['getAllCubeCombos', 'generateMaterialCombos', 'generateRExtraCubes', 'getTheoreticalReflectances']

MCombo = Tuple[Material, Material]  # This is just an alias to shorten some of the type hinting.


class CubeCombo:
    """A convenient way of packaging together two ImCubes along with the material that each ImCube is an image of.

    Args:
        material1: The material that was imaged in `cube1`
        material2: The material that was imaged in `cube2`
        cube1: An ImCube with an image of a glass-{material} interface where the material is `material1`.
        cube2: An ImCube with an image of a glass-{material} interface where the material is `material2`.

    """
    def __init__(self, material1: Material, material2: Material, cube1: ImCube, cube2: ImCube):
        self.mat1 = material1
        self.mat2 = material2
        self.data1 = cube1
        self.data2 = cube2

    def keys(self) -> MCombo:
        return self.mat1, self.mat2

    def values(self) -> Tuple[ImCube, ImCube]:
        return self.data1, self.data2

    def items(self) -> Iterator[Tuple[Material, ImCube]]:
        return zip(self.keys(), self.values())

    def __getitem__(self, item: Material) -> ImCube:
        if item == self.mat1:
            return self.data1
        elif item == self.mat2:
            return self.data2
        else:
            raise KeyError("Key must be either mat1 or mat2.")


@dataclass
class _ComboSummary:
    """A convenient packaging of information related to the extra reflectance calculated from a single _CubeCombo"""
    mat1Spectra: np.ndarray
    mat2Spectra: np.ndarray
    combo: CubeCombo
    rExtra: np.ndarray
    I0: np.ndarray
    cFactor: float
    weight: np.ndarray


def _interpolateNans(arr):
    """Interpolate out nan values along the third axis of an array"""

    def interp1(arr1):
        nans = np.isnan(arr1)
        f = lambda z: z.nonzero()[0]
        arr1[nans] = np.interp(f(nans), f(~nans), arr1[~nans])
        return arr1

    arr = np.apply_along_axis(interp1, 2, arr)
    return arr


def getTheoreticalReflectances(materials: Set[Material], wavelengths: Tuple[float], numericalAperture: float) -> Dict[Material, pd.Series]:
    """Generate a dictionary containing a pandas series of the `material`-glass reflectance for each material in
    `materials`.

    Args:
        materials: The set of materials that you want to retrieve the theoretical reflectance for.
        wavelengths: The wavelengths that you want the reflectances calculated at.
        numericalAperture: The numerical aperture that the reflectance should be calculated at.

    Returns:
        A dictionary of the reflectances for each material.
    """

    theoryR = {}
    for material in materials:  # For each unique material in the `cubes` list
        logging.getLogger(__name__).info(f"Calculating reflectance for {material}")
        theoryR[material] = reflectanceHelper.getReflectance(material, Material.Glass, wavelengths=wavelengths, NA=numericalAperture)
    return theoryR


def generateMaterialCombos(materials: Iterable[Material], excludedCombos: Optional[Iterable[MCombo]] = None) -> List[MCombo]:
    """Given a list of materials, this function returns a list of all possible material combo tuples.

    Args:
        materials: The list of materials that you want to generate every possible combo of.
        excludedCombos: Combinations of materials that you don't want included in the combinations.

    Returns:
        A list of Material combinations.

    """
    if excludedCombos is None:
        excludedCombos = []
    matCombos = list(itertools.combinations(materials, 2))  # All the combinations of materials that can be compared
    matCombos = [(m1, m2) for m1, m2 in matCombos if
                 not (((m1, m2) in excludedCombos) or ((m2, m1) in excludedCombos))]  # Remove excluded combinations.
    for i, (m1, m2) in enumerate(matCombos):  # Make sure to arrange materials so that our reflectance ratio is greater than 1
        if (reflectanceHelper.getReflectance(m1, Material.Glass) / reflectanceHelper.getReflectance(m2, Material.Glass)).mean() < 1:
            matCombos[i] = (m2, m1)
    return matCombos


def getAllCubeCombos(matCombos: Iterable[MCombo], df: pd.DataFrame) -> Dict[MCombo, List[CubeCombo]]:
    """Given a list of material combo tuples, return a dictionary whose keys are the material combo tuples and whose values are
    lists of CubeCombos.

    Args:
        matCombos: A list of material combinations, most likely generated by `generateMaterialCombos`
        df: A Pandas DataFrame containing a 'material' column of the materials, and a 'cube' column containing the associated ImCube

    Returns:
        A dictionary with a key for each material combination. Each value is a list of all the `CubeCombo`s extracted from `df`.
    """
    allCombos = {}
    for matCombo in matCombos:
        matCubes = {material: df[df['material'] == material]['cube'] for material in matCombo}  # A dictionary sorted by material containing lists of the ImCubes that are relevant to this loop..
        allCombos[matCombo] = [CubeCombo(*matCubes.keys(), *combo) for combo in itertools.product(*matCubes.values())]
    allCombos = {key: val for key, val in allCombos.items() if len(val) > 0}  # In some cases a matCombo appears for which there is no data. get rid of these.
    return allCombos


def _calculateSpectraFromCombos(cubeCombos: Dict[MCombo, List[CubeCombo]], theoryR: Dict[Material, pd.Series],
                                numericalAperture: float, mask: Optional[Roi] = None) ->\
        Tuple[Dict[Union[MCombo, str], Dict[str, Any]], Dict[MCombo, List[_ComboSummary]]]:
    """This is used to examine the output of extra reflection calculation before using saveRExtra to save a cube for each setting.
    Expects a dictionary as created by `getAllCubeCombos` and a dictionary of theoretical reflections.

    Args:
        cubeCombos: A dictionary containing all possible combinations of ImCubes measured at different materials. Keyed by the material combo.
        theoryR: A dictionary containing the theoretical reflectances for all `Materials` in use. Should be accurate for the `numericalAperture` in question.
        numericalAperture: The illumination numerial aperture that the images were taken at.
        mask: An ROI that limits the region of the ImCubes that is analyzed. The spectra will be averaged over this region. If `None` the spectra will be average over the full XY FOV.

    Returns:
        The first item is a dictionary containing information about the average calculation for each material combination
        as well as the average calculation accross all material combinations. The seconds item is a dictionary containing
        information about every single cube combo.
    """

    # Save the results of relevant calculations to a dictionary, this dictionary will be returned to the user along with
    # the raw data, `allCombos`
    allCombos = {}
    meanValues = {}
    params = ['rExtra', 'I0', 'mat1Spectra', 'mat2Spectra']
    for matCombo in cubeCombos.keys():
        allCombos[matCombo] = []
        for combo in cubeCombos[matCombo]:
            mat1, mat2 = combo.keys()
            c = _ComboSummary(mat1Spectra=combo[mat1].getMeanSpectra(mask)[0],
                              mat2Spectra=combo[mat2].getMeanSpectra(mask)[0],
                              weight=None,
                              rExtra=None,
                              I0=None,
                              cFactor=None,
                              combo=combo)
            c.weight = (c.mat1Spectra - c.mat2Spectra) ** 2 / (c.mat1Spectra ** 2 + c.mat2Spectra ** 2)
            c.rExtra = ((theoryR[mat1] * c.mat2Spectra) - (theoryR[mat2] * c.mat1Spectra)) / (c.mat1Spectra - c.mat2Spectra)
            c.I0 = c.mat2Spectra / (theoryR[mat2] + c.rExtra)
            waterTheory = reflectanceHelper.getReflectance(Material.Water, Material.Glass, wavelengths=list(theoryR.values())[0].index, NA=numericalAperture)
            c.cFactor = (c.rExtra.mean() + waterTheory.mean()) / waterTheory.mean()
            allCombos[matCombo].append(c)
        meanValues[matCombo] = {}
        for param in params:
            meanValues[matCombo][param] = np.average(np.array(list([getattr(combo, param) for combo in allCombos[matCombo]])),
                                                    axis=0,
                                                    weights=np.array([combo.weight for combo in allCombos[matCombo]]))
        meanValues[matCombo]['cFactor'] = np.average(np.array([combo.cFactor for combo in allCombos[matCombo]]),
                                                     axis=0,
                                                     weights=np.array([combo.weight.mean() for combo in allCombos[matCombo]]))
        meanValues[matCombo]['weight'] = np.mean(np.array([combo.weight for combo in allCombos[matCombo]]))
    meanValues['mean'] = {param: np.average(np.array(list([meanValues[matCombo][param] for matCombo in cubeCombos.keys()])),
                                            axis=0,
                                            weights=np.array([meanValues[matCombo]['weight'] for matCombo in cubeCombos.keys()])) for param in params}
    meanValues['mean']['cFactor'] = np.average(np.array([meanValues[matCombo]['cFactor'] for matCombo in cubeCombos.keys()]),
                                                 axis=0,
                                                 weights=np.array([meanValues[matCombo]['weight'].mean() for matCombo in cubeCombos.keys()]))
    return meanValues, allCombos


def plotExtraReflection(df: pd.DataFrame, theoryR: Dict[Material, pd.Series], matCombos: List[MCombo],
                        numericalAperture: float, mask: Optional[Roi] = None,
                        plotReflectionImages: bool = False) -> List[plt.Figure]:
    """Generate a variety of plots displaying information about the extra reflectance calculation.
    
    Args:
        df: A pandas dataframe containing a row for each ImCube that is to be included in the calculation. The dataframe
            should have the following columns: 'cube': The `ImCube` object in question, should be an image of a
            glass-{material} interface. `material`: The `Material` that the ImCube is an image of. `setting`: A string
            describing the imaging configuration. If the dataframe has multiple settings then each settings will be
            processed separately and compared.
        theoryR: A dictionary where the key is a `Material` and the value is a pandas series giving the reflectance for
            a glass-{material} reflection over a range of wavelengths. The index of the series should be the wavelengths.
        matCombos: A list of the various material combinations that should be evaluated.
        numericalAperture: The numerical aperture that the ImCubes being used were imaged at.
        mask: An ROI indicating the region of the images that should be included in the evaluation.
        plotReflectionImages: An optional parameter. If True additional plots will be opened.

    Returns:
        A list of matplotlib figures resulting from this calculation.

    """
    settings = set(df['setting'])
    meanValues: Dict[str, Dict[Union[MCombo, str], Dict[str, Any]]] = {}
    allCombos: Dict[str, Dict[MCombo, List[_ComboSummary]]] = {}
    for sett in settings:
        cubeCombos = getAllCubeCombos(matCombos, df[df['setting'] == sett])
        meanValues[sett], allCombos[sett] = _calculateSpectraFromCombos(cubeCombos, theoryR, numericalAperture, mask)
    dock = DockablePlotWindow("Primary")
    figs = [dock]
    fig, ax = dock.subplots("Extra Reflection")  # For extra reflections
    ax.set_ylabel("Reflectance (1 = total reflection)")
    ax.set_xlabel("nm")
    i = 0
    numLines = []
    for sett in settings:
        for matCombo in allCombos[sett].keys():
            numLines.append(len(allCombos[sett][matCombo]))
    numLines = sum(numLines)
    colormap = plt.cm.gist_rainbow
    colors = [colormap(i) for i in np.linspace(0, 0.99, numLines)]
    for sett in settings:
        for matCombo in allCombos[sett].keys():
            mat1, mat2 = matCombo
            for combo in allCombos[sett][matCombo]:
                cubes = combo.combo
                ax.plot(cubes[mat1].wavelengths, combo.rExtra, color=colors[i],
                        label=f'{sett} {mat1}:{int(cubes[mat1].metadata.exposure)}ms {mat2}:{int(cubes[mat2].metadata.exposure)}ms')
                i += 1
        ax.plot(cubes[mat1].wavelengths, meanValues[sett]['mean']['rExtra'], color='k', label=f'{sett} mean')  # TODO Add a hover annotation since all of the lines are black it's impossible to know which one is which.
    ax.legend()

    fig2, ratioAxes = dock.subplots("Reflectance Ratios", nrows=len(matCombos))  # for correction factor
    if not isinstance(ratioAxes, np.ndarray):
        ratioAxes = np.array(ratioAxes).reshape(1)  # If there is only one axis we still want it to be a list for the rest of the code
    ratioAxes = dict(zip(matCombos, ratioAxes))
    for combo in matCombos:
        ratioAxes[combo].set_title(f'{combo[0].name}/{combo[1].name} reflection ratio')
        ratioAxes[combo].plot(theoryR[combo[0]] / theoryR[combo[1]], label='Theory')
    for sett in settings:
        for matCombo in allCombos[sett].keys():
            mat1, mat2 = matCombo
            for combo in allCombos[sett][matCombo]:
                cubes = combo.combo
                ratioAxes[matCombo].plot(cubes[mat1].wavelengths, combo.mat1Spectra / combo.mat2Spectra,
                                         label=f'{sett} {mat1.name}:{int(cubes[mat1].metadata.exposure)}ms {mat2.name}:{int(cubes[mat2].metadata.exposure)}ms')
    [ratioAxes[combo].legend() for combo in matCombos]

    dock = DockablePlotWindow("Correction Comparison")
    figs.append(dock)
    for sett in settings:
        settMatCombos = allCombos[sett].keys()  # Sometime we are looking at settings which don't have all the same matCombos. Only use the combos specific to this setting.
        means = meanValues[sett]['mean']
        fig6, scatterAx3 = dock.subplots(f"{sett}")
        scatterAx3.set_ylabel("Theoretical Ratio")
        scatterAx3.set_xlabel("Observed Ratio. No correction")
        scatterPointsY = [(theoryR[matCombo[0]] / theoryR[matCombo[1]]).mean() for matCombo in settMatCombos]
        scatterPointsX = [(meanValues[sett][matCombo]['mat1Spectra'] / meanValues[sett][matCombo]['mat2Spectra']).mean() for
                          matCombo in settMatCombos]
        [scatterAx3.scatter(x, y, label=f'{matCombo[0].name}/{matCombo[1].name}') for x, y, matCombo in zip(scatterPointsX, scatterPointsY, settMatCombos)]
        x = np.array([0, max(scatterPointsX)])
        scatterAx3.plot(x, x, label='1:1')
        scatterAx3.legend()

        fig3, scatterAx = dock.subplots(f'{sett} cFactor: {means["cFactor"]}')  # A scatter plot of the theoretical vs observed reflectance ratio.
        scatterAx.set_ylabel("Theoretical Ratio")
        scatterAx.set_xlabel("Observed Ratio w/ cFactor")
        scatterPointsY = [(theoryR[matCombo[0]] / theoryR[matCombo[1]]).mean() for matCombo in settMatCombos]
        scatterPointsX = [means['cFactor'] * (
                meanValues[sett][matCombo]['mat1Spectra'] / meanValues[sett][matCombo]['mat2Spectra']).mean() for
                          matCombo in settMatCombos]
        [scatterAx.scatter(x, y, label=f'{matCombo[0].name}/{matCombo[1].name}') for x, y, matCombo in zip(scatterPointsX, scatterPointsY, settMatCombos)]
        x = np.array([0, max(scatterPointsX)])
        scatterAx.plot(x, x, label='1:1')
        scatterAx.legend()

        fig4, scatterAx2 = dock.subplots(sett)  # A scatter plot of the theoretical vs observed reflectance ratio.
        scatterAx2.set_ylabel("Theoretical Ratio")
        scatterAx2.set_xlabel("Observed Ratio after Subtraction")
        scatterPointsY = [(theoryR[matCombo[0]] / theoryR[matCombo[1]]).mean() for matCombo in settMatCombos]
        scatterPointsX = [np.nanmean((meanValues[sett][matCombo]['mat1Spectra'] - means['I0'] * means['rExtra']) / (
                meanValues[sett][matCombo]['mat2Spectra'] - means['I0'] * means['rExtra'])) for matCombo in
                          settMatCombos]
        [scatterAx2.scatter(x, y, label=f'{matCombo[0].name}/{matCombo[1].name}') for x, y, matCombo in
         zip(scatterPointsX, scatterPointsY, settMatCombos)]
        x = np.array([0, max(scatterPointsX)])
        scatterAx2.plot(x, x, label='1:1')
        scatterAx2.legend()

        if plotReflectionImages:
            dock = DockablePlotWindow("System Reflectance% Estimates")
            figs.append(dock)
            for matCombo in settMatCombos:
                mat1, mat2 = matCombo
                for combo in allCombos[sett][matCombo]:
                    cubes = combo.combo
                    fig, ax = dock.subplots(f"{sett}, {mat1}:{int(cubes[mat1].metadata.exposure)}ms, {mat2}:{int(cubes[mat2].metadata.exposure)}ms")
                    _ = ((np.array(theoryR[mat1])[np.newaxis, np.newaxis, :] * cubes[mat2].data) - (
                        np.array(theoryR[mat2])[np.newaxis, np.newaxis, :] * cubes[mat1].data)) / (
                            cubes[mat1].data - cubes[mat2].data)
                    _[np.isinf(_)] = np.nan
                    if np.any(np.isnan(_)):
                        _ = _interpolateNans(_)  # any division error resulting in an inf will really mess up our refIm. so we interpolate them out.
                    refIm = _.mean(axis=2)
                    im = ax.imshow(refIm, vmin=np.percentile(refIm, .5), vmax=np.percentile(refIm, 99.5))
                    plt.colorbar(mappable=im)

        logging.getLogger(__name__).info(f"{sett} correction factor")
    return figs


def _generateOneRExtraCube(combo: CubeCombo, theoryR: Dict[Material, pd.Series]) -> Tuple[np.ndarray, np.ndarray]:
    """Given a combination of two ImCubes imaging different materials and the theoretical reflectance for the materials
    imaged in the two ImCubes this function will generate an estimation of the extra reflectance in the system.

    Args:
        combo: A combo of two ImCubes imaging two different glass-material interfaces.
        theoryR: A dictionary providing pandas series giving the theoretically expected reflectance for each glass-material
            interface represented by the two ImCubes.

    Returns:
        The first item is the array of the estimated extra reflectance of the system, expressed as values between 0 and
        1 (this is the same shape as the ImCubes supplied to the function.) The second item is an estimate of how reliable
        each element in the array is.

    """
    data1 = combo.data1.data
    data2 = combo.data2.data
    T1 = np.array(theoryR[combo.mat1])[np.newaxis, np.newaxis, :]
    T2 = np.array(theoryR[combo.mat2])[np.newaxis, np.newaxis, :]
    denominator = data1 - data2
    nominator = T1 * data2 - T2 * data1
    arr = nominator / denominator

    #Even when a weight is calculated as zero if we have weird values (np.nan, np.inf) in arr we will get a messed up end result.
    arr[np.isinf(arr)] = 0
    arr[np.isnan(arr)] = 0
    arr[arr < 0] = 0
    arr[arr > 1] = 1
    #calculate a confidence weighting for every point in the cube.
    # According to propagation of error if we assume that TheoryR has no error
    # and the data (camera counts) has a constant error of C then the error is C * sqrt((T1-T2)*data1^2 + (T2-T1)*data2^2) / (data1 - data2)^2
    # Since we are just looking for a relative measure of confidence we can ignore C. We use the `Variance weighted average'
    # (1/stddev^2)
    #Doing this calculation with noise in Theory instead of data gives us a variance of C^2 * (data1^2 + data2^2) / (data1 - data2)^2. this seems like a better equation to use. TODO Really? Why?
    weight = (data1-data2)**2 / (data1**2 + data2**2) # The weight is the inverse of the variance. Higher weight = more reliable data.
    return arr, weight


def generateRExtraCubes(allCombos: Dict[MCombo, List[CubeCombo]], theoryR: dict, numericalAperture: float) -> \
        Tuple[ExtraReflectanceCube, Dict[Union[str, MCombo], Tuple[np.ndarray, np.ndarray]]]:
    """Generate a series of extra reflectance cubes based on the input data.

    Args:
        allCombos: a dict of lists CubeCombos, each keyed by a 2-tuple of Materials.
        theoryR: the theoretical reflectance for each material.
        numericalAperture: The numerical aperture that the ImCubes were imaged at. The theoryR reflectances should have
            also been calculated at this NA

    Returns:
        Returns extra reflectance for each material combo as well as the mean of all extra reflectances.
        This is what gets used. Ideally all the cubes will be very similar.
        Additionally returns a list of plot objects. references to these must be kept alive for the plots to be responsive.
    """
    rExtra = {}
    for matCombo, combosList in allCombos.items():
        logging.getLogger(__name__).info(f"Calculating rExtra for: {matCombo}")
        erCubes, weights = zip(*[_generateOneRExtraCube(combo, theoryR) for combo in combosList])
        weightSum = reduce(lambda x, y: x + y, weights)
        weightedMean = reduce(lambda x, y: x + y, [cube*weight for cube, weight in zip(erCubes, weights)]) / weightSum
        meanWeight = weightSum / len(weights)
        rExtra[matCombo] = (weightedMean, meanWeight)
    erCubes, weights = zip(*rExtra.values())
    weightSum = reduce(lambda x, y: x + y, weights)
    weightedMean = reduce(lambda x, y: x + y, [cube * weight for cube, weight in zip(erCubes, weights)]) / weightSum
    meanWeight = weightSum / len(weights)
    rExtra['mean'] = (weightedMean, meanWeight)
    sampleCube: ImCube = list(allCombos.values())[0][0].data1
    md = ERMetaData(sampleCube.metadata.dict, numericalAperture)
    erCube = ExtraReflectanceCube(rExtra['mean'][0], sampleCube.wavelengths, md)
    return erCube, rExtra






# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
An adaptation of the PWS analysis script used in MATLAB. This is now outdated. The up to date version is in
`pwspy.analysis.pws`.
"""
import logging

from pwspy.dataTypes import ImCube, KCube
import copy
import scipy.signal as sps
import numpy as np

from pwspy.analysis.pws import PWSAnalysisResults


def analyzeCube(cubeCell: ImCube, darkCount: int, mirror: ImCube, orderFilter: int,
                cutoffFilter: float, wavelengthStart: int, wavelengthStop: int,
                orderPolyFit: int, isAutocorrMinSub: bool, indexAutocorrLinear: int,
                isOpdPolysub: bool, isHannWindow: bool):
    # Indicate the OPD Stop Index.
    indexOpdStop = 100

    logger = logging.getLogger(__name__)

    cube = copy.deepcopy(cubeCell)  # We don't want to mess up the original cube.

    logger.info("Normalizing ImCubes")
    cube.subtractDarkCounts(darkCount)
    cube.normalizeByExposure()
    mirror.subtractDarkCounts(darkCount)
    mirror.normalizeByExposure()
    cube = cube / mirror

    logger.info("Filtering Signal")
    b, a = sps.butter(orderFilter,
                      cutoffFilter)  # The cutoff totally ignores what the `sample rate` is. so a 2nm interval image cube will be filtered differently than a 1nm interval cube. This is how it is in matlab.
    cube.data = sps.filtfilt(b, a, cube.data, axis=2)

    # The rest of the analysis will be performed only on the selected wavelength range.
    cube = cube.selIndex(wavelengthStart, wavelengthStop)

    # Determine the mean-reflectance for each pixel in the cell.
    reflectance = cube.data.mean(axis=2)

    ## -- Convert to K-Space
    logger.info("Converting to K-Space")
    cube = KCube.fromImCube(cube)

    ## -- Polynomial Fit
    logger.info("Subtracting Polynomial")
    flattenedData = cube.data.reshape((cube.data.shape[0] * cube.data.shape[1], cube.data.shape[2]))
    flattenedData = np.rollaxis(flattenedData, 1)  # Flatten the array to 2d and put the wavenumber axis first.
    cubePoly = np.zeros(flattenedData.shape)  # make an empty array to hold the fit values.
    polydata = np.polyfit(cube.wavenumbers, flattenedData,
                          orderPolyFit)  # At this point polydata goes from holding the cube data to holding the polynomial values for each pixel. still 2d.
    for i in range(orderPolyFit + 1):
        cubePoly += (np.array(cube.wavenumbers)[:, np.newaxis] ** i) * polydata[i,
                                                                       :]  # Populate cubePoly with the fit values.
    cubePoly = np.moveaxis(cubePoly, 0, 1)
    cubePoly = cubePoly.reshape(cube.data.shape)  # reshape back to a cube.

    # Remove the polynomial fit from filtered cubeCell.
    cube.data = cube.data - cubePoly

    ## RMS - POLYFIT
    # The RMS should be calculated on the mean-subtracted polyfit. This may
    # also be accomplished by calculating the standard-deviation.
    rmsPoly = cubePoly.std(axis=2)

    ## -- RMS
    # Obtain the RMS of each signal in the cube.
    rms = cube.data.std(axis=2)

    ## -- Autocorrelation
    logger.info("Calculating Autocorrelation")
    slope, rSquared = cube.getAutoCorrelation(isAutocorrMinSub, indexAutocorrLinear)

    ## OPD Analysis
    logger.info("Calculating OPD")
    if isOpdPolysub:  # If cubeOpdPolysub is to be generated
        opd, xvalOpd = cube.getOpd(isHannWindow, indexOpdStop)
    else:
        opd = xvalOpd = None

    ## Ld Calculation
    k = 2 * np.pi / 0.55
    fact = 1.38 * 1.38 / 2 / k / k
    A1 = 0.008
    A2 = 4
    ld = ((A2 / A1) * fact) * (rms / (-1 * slope.reshape(rms.shape)))
    ## Outputs
    results = PWSAnalysisResults.create(
        reflectance=reflectance,
        rms=rms,
        polynomialRms=rmsPoly,
        autoCorrelationSlope=slope,
        rSquared=rSquared,
        ld=ld,
        opd=opd,
        opdIndex=xvalOpd)
    return results


if __name__ == '__main__':
    cube = ImCube.loadAny(r'G:\Calibrations\CellPhantom\lcpws1\5th\Cell1')
    mirror = ImCube.loadAny(r'G:\Calibrations\CellPhantom\lcpws1\5th\Cell666')
    # Default settings
    darkCounts = 1957
    results = analyzeCube(cube, darkCounts, mirror, 6, 0.15, 510, 690, 0, True, 7, True, True)



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
This script blurs an image cube in the xy direction. Allows you to turn an
image of cells into something that can be used as a reference image, assuming
most of the the FOV is glass.
"""

import copy

import matplotlib.pyplot as plt
from pwspy.dataTypes import ImCube


if __name__ == '__main__':
    a = ImCube.fromTiff(r'G:\Vasundhara_MSCs_data\Cell1170')

    mirror = copy.deepcopy(a)  # This doesn't work right. maybe becuaes of the wi
    mirror.filterDust(10)

    a.plotMean()
    mirror.plotMean()
    norm = (a / mirror)
    norm.plotMean()
    plt.figure()
    plt.imshow(norm.data.std(axis=2))



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""An adaptation of Adam Eshein's (aeshein@u.northwestern.edu) MATLAB code for converting PWS
RMS measurements to D. The MATLAB code was based on Vadim-Backman's (v-backman@northwestern.edu) original code in
Mathematica.

References:
   L. Cherkezyan, D. Zhang, H. Subramanian, I. Capoglu, A. Taflove,
   V. Backman, "Review of interferometric spectroscopy of scattered light
   for the quantification of subdiffractional structure of biomaterials."
   J. of Biomedical Optics, 22(3), 030901 (2017).

Primary Functions
-----------------
.. autosummary::
   :toctree: generated/

   sigma2D
   sigma2DApproximation

Secondary Functions
--------------------
These functions are called by the primary functions and probably don't need to be used directly.

.. autosummary::
   :toctree: generated/

   expn
   acf
   acfd
   calcDSize

"""
import typing
import numpy as np
from scipy.integrate import quad

NumberOrArray = typing.Union[np.ndarray, float]

@np.vectorize
def expn(n: float, x: float) -> float:
    """
    This is a custom implementation of the generalized exponential exponential integral. The implementation in `scipy.special.expn` is faster but only supports
    integers for `n`.

    :math:`\\int_0^{\\infty} \\! \\frac{e^{-x*t}}{t^n} \\, \\mathrm{d}t`


    Args:
        n: The exponent of the divisor
        x: The exponent of the inverse exponential

    Returns:
        The result of the integral :math:`\\int_0^{\\infty} \\! \\frac{e^{-x*t}}{t^n} \\, \\mathrm{d}t`
    """
    def integrand(t: float, N: float, X: float):
        return np.exp(-X*t) / t**N
    return quad(integrand, 1, np.inf, args=(n, x))[0]


def acf(d: NumberOrArray, lmin: NumberOrArray, lmax: NumberOrArray, x: NumberOrArray) -> NumberOrArray:
    """This function is based on the `acf_1` MATLAB function.

    Args:
        d: :todo:
        lmin: :todo:
        lmax: :todo:
        x: :todo:

    Returns:

    """
    out = ((3 - d) * ((((lmin**4) * ((lmin / lmax)**(-d)) * expn(-2 + d, x / lmax)) / (lmax**3)) - lmin * expn(-2 + d, x / lmin))) / (lmin * (1 - (lmin / lmax)**(3-d)))
    return out


def acfd(d: NumberOrArray, lmin: NumberOrArray, lmax: NumberOrArray) -> NumberOrArray:
    """This function is based on the `acfd` MATLAB function. The equation has been algebraically refactored from the
    MATLAB code to require fewer computations of logarithms.

    Args:
        d: :todo:
        lmin: :todo:
        lmax: :todo:

    Returns:
        :todo:
    """
    delta = 0.1
    x = (lmax+lmin) / 100
    out = 3 + np.log(acf(d, lmin, lmax, x + delta) / acf(d, lmin, lmax, x)) / np.log((x + delta) / x)
    return out


def calcDSize(raw_rms: np.ndarray, noise: float, NAi: float):
    """
    The `system_correction` argument from the original MATLAB function has been excluded. We assume that any RMS
    values being provided to this function have already been properly corrected for hardware defects.

    Args:
        raw_rms: An array of RMS values you wish to convert.
        noise: A positive value indicating the background noise in the system in terms of RMS (E.G. the RMS of
            a measurement of clean glass). On the Backman lab "STORM" system this has been previously
            found to be 0.012. For the "LCPWS1" system it has been found to be 0.009.
        NAi: The illumination numerical aperture (NA) of the objective.

    Returns:
        d_size: :todo:
    """
    sigma = np.sqrt(raw_rms**2 - noise**2)  # The MATLAB version will return complex numbers if this is negative. This implementation returns `nan`
    d_size = sigma * 13.8738 * NAi + 1.473
    return d_size


def sigma2D(raw_rms: np.ndarray, noise: float, NAi: float) -> np.ndarray:
    """Converts d_size to D precisely. This function can be significantly slower than
    `sigma2DApproximation` but more closely follows the analytical solution.

    Args:
        raw_rms: An array of RMS values you wish to convert.
        noise: A positive value indicating the background noise in the system in terms of RMS (E.G. the RMS of
            a measurement of clean glass). On the Backman lab "STORM" system this has been previously
            found to be 0.012. For the "LCPWS1" system it has been found to be 0.009.
        NAi: The illumination numerical aperture (NA) of the objective.

    Returns:
        Results from the exact solution of converting from Sigma to D. This calculation is based on
        derivations by Vadim Backman.
    """
    d_size = calcDSize(raw_rms, noise, NAi)
    #This runs 6 times faster than the matlab version for some reason
    mf = 1000000
    d_size[d_size == 3] = 3.00001  # d_size == 3 will result in a singularity so we nudge the data a bit.

    lmaxlminapprox = 100
    correction = ((3 - d_size) * (1 - (lmaxlminapprox**(-1. * d_size)))) / (
                d_size * (1 - (lmaxlminapprox**(d_size - 3))))
    mass = mf / correction
    d_exact = acfd(d_size, 1, mass ** (1. / d_size))
    return d_exact


def sigma2DApproximation(raw_rms: np.ndarray, noise: float, NAi: float) -> np.ndarray:
    """Converts d_size to D using a 15th degree polynomial approximation of the `sigma2D` function. This function can be
    significantly faster than `sigma2D` but is not guaranteed to be accurate especially at extreme input values.

    Args:
        raw_rms: An array of RMS values you wish to convert.
        noise: A positive value indicating the background noise in the system in terms of RMS (E.G. the RMS of
            a measurement of clean glass). On the Backman lab "STORM" system this has been previously
            found to be 0.012. For the "LCPWS1" system it has been found to be 0.009.
        NAi: The illumination numerical aperture (NA) of the objective.

    Returns:
        An estimation of D calculated from Sigma. It's based on a 15th order polynomial fit of
        the output from the `sigma2D` function. No value of D yields error >0.1%.
    """
    d_size = calcDSize(raw_rms, noise, NAi)
    # These are the polynomical coefficients stored in the SimgaToD_coefs.mat file. They go from high order to low, E.G. x^2 + x + 1
    sigmaToD_coefs = [-9.14414809736752e-09, 8.02336561707375e-07, -3.22276589702395e-05, 0.000784980326922923, -0.0129458554989855, 0.152852475387947,
                      -1.33210715342735, 8.70614624955508, -42.9149123685218, 159.111116839950, -438.829185621276, 881.674160348790, -1246.22822358504,
                      1168.11294529161, -647.810667596662, 161.021813781994]
    sigmaToD_coefs = np.array(list(reversed(sigmaToD_coefs))) # The order of the coefficients is different than in matlab
    d_estimate = np.polynomial.polynomial.polyval(d_size, sigmaToD_coefs)
    d_estimate[d_size > 10] = 2.99  # The fitting doesn't work well at very high values of D_size.
    return d_estimate


if __name__ == '__main__':
    import time
    testData = np.linspace(0.001, 0.1, num=100)
    stime = time.time()
    exact = sigma2D(testData, 0.009, 0.55)
    estimate = sigma2DApproximation(testData, 0.009, 0.55)
    ftime = time.time() - stime  # Execution time of the function
    a = 1



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
Plots the spectra of the ITO thin film used for calibration.
"""
from pwspy.dataTypes import ImCube, AcqDir
from glob import glob
import matplotlib.pyplot as plt
import numpy as np
import os
from datetime import datetime
from pwspy import dateTimeFormat
import matplotlib as mpl
from pwspy.utility.fileIO import loadAndProcess


def proc(im: ImCube):
    """This function is run in a separate thread/process to load the files."""
    refPath = os.path.join(os.path.split(im.metadata.acquisitionDirectory.filePath)[0], 'Cell999')
    ref = ImCube.fromMetadata(AcqDir(refPath).pws)
    ref.correctCameraEffects()
    im.correctCameraEffects()
    ref.normalizeByExposure()
    im.normalizeByExposure()
    return im, ref

if __name__ == '__main__':
    subDir = r'J:\Calibrations\ITOThinFilm\LCPWS1' # Use '*' to include all subdirectories
    dateStart = '1-16-2020'
    dateEnd = None
    templateName = '1_16_2020'
    
    files = glob(os.path.join(subDir, '*'))
    dates = [os.path.split(f)[-1] for f in files]
    
    
    
    ims = []
    for i in files:
        try:
            ims.append(AcqDir(os.path.join(i, 'Cell1')).pws)
        except:
            print("Skip ", i)
    ims = loadAndProcess(ims, processorFunc=proc)
    
    if dateStart is not None: dateStart = datetime.strptime(dateStart, '%m-%d-%Y')
    else: dateStart = datetime.min
    if dateEnd is not None: dateEnd = datetime.strptime(dateEnd, '%m-%d-%Y')
    else: dateEnd = datetime.max
    
    for i, ref in ims:
        try:
            i.date = datetime.strptime(i.metadata.time, dateTimeFormat)
        except:
            i.date = datetime.strptime(i.metadata.time, dateTimeFormat)
        i.dateString = datetime.strftime(i.date, '%m-%d-%Y')
    
    files = [f for (im, ref), f in zip(ims,files) if dateStart < im.date < dateEnd]
    ims = [(im, ref) for im, ref in ims if dateStart < im.date < dateEnd]
    dates = [datetime.strptime(im.metadata.time, dateTimeFormat) for im, ref in ims]
    dates, ims, files = zip(*sorted(zip(dates, ims, files)))
    
    template = None
    for f, (im, ref) in zip(files,ims):
        if templateName in os.path.split(f):
            print('getting template spectra')
            template = im.getMeanSpectra()[0] / ref.getMeanSpectra()[0]
                
    if template is None:
        print("No template found.")
    
    
    fig, ax = plt.subplots()
    fig.suptitle("Raw Spectrum")
    fig2,ax2 = plt.subplots()
    fig2.suptitle("Mirror Spectrum")
    fig3, ax3 = plt.subplots()
    fig3.suptitle("Normalized Spectrum")
    fig4, ax4 = plt.subplots()
    fig4.suptitle('Template Correlation')
    corrs=[]
    
    colors = mpl.cm.get_cmap('gist_rainbow')(np.linspace(0,1,num=len(files)))
    
    for i, ((im, ref), file) in enumerate(zip(ims, files)):
        #plot spectra
        spectra = im.getMeanSpectra()
        ax.plot(im.wavelengths, spectra[0],label = im.dateString, color = colors[i])
        ax.fill_between(im.wavelengths, spectra[0] + spectra[1], spectra[0] - spectra[1], alpha = 0.4, color = colors[i])
        #Plot mirror spectra
        refSpectra = ref.getMeanSpectra()
        ax2.plot(im.wavelengths, refSpectra[0],label = im.dateString, color = colors[i])
        ax2.fill_between(im.wavelengths, refSpectra[0] + refSpectra[1], refSpectra[0] - refSpectra[1], alpha = 0.4, color = colors[i])
        #plot normalized spectra
        ax3.plot(im.wavelengths, spectra[0] / refSpectra[0],label = im.dateString, color = colors[i])
        std = np.sqrt(spectra[0]**2 / refSpectra[0]**2 * (spectra[1]**2 / spectra[0]**2 + refSpectra[1]**2 / refSpectra[0]**2), )
    #    ax3.fill_between(im.wavelengths,spectra[0] / refSpectra[0] + std, spectra[0] / refSpectra[0] - std, alpha = 0.4, color = colors[i])
        #Calculate correlation
        if template is not None:
            a, b = (spectra[0]/refSpectra[0], template)
            a = (a - np.mean(a)) / (np.std(a) * len(a))
            b = (b - np.mean(b)) / (np.std(b))
            corrs.append(np.correlate(a, b)[0])
    ax.legend()
    ax2.legend()
    ax3.legend()
        
    ax4.bar([i.dateString for i, ref in ims], corrs)
    ax4.set_ylim([.7, 1])
    [tick.set_rotation(45) for tick in ax4.get_xticklabels()]



# -*- coding: utf-8 -*-
"""
Loop through all ROIs for all acquisitions in a directory and plot a histogram of the RMS values within the ROI.

@author: Nick Anthony
"""

if __name__ == '__main__':

    import pwspy.dataTypes as pwsdt
    import matplotlib.pyplot as plt
    import pathlib
    import numpy as np

    workingDirectory = r'C:\Users\nicke\Desktop\cells'  # The folder that all your acquisitions are saved under.
    analysisName = 'script'  # This will often be "p0"


    def plotHist(roi, rms):
        """
        This function takes an ROI
        and a 2D RMS image and plots a histogram of the RMS values inside the ROI
        """
        # Check input values just to be safe.
        assert isinstance(roi, pwsdt.Roi)  # Make sure roiFile variable is actually an ROI
        assert isinstance(rms, np.ndarray)  # Make sure the RMS image is an numpy array
        assert roi.mask.shape == rms.shape  # Make sure the ROI and RMS arrays have the same dimensions.

        vals = rms[roi.mask]  # A 1D array of the values inside the ROI
        plt.hist(vals)  # Plot a histogram


    cellFolderIterator = pathlib.Path(workingDirectory).glob("Cell[0-9]")  # An iterator for all folders that are below workingDirectory and match the "regex" pattern "Cell[0-9]"
    for folder in cellFolderIterator:
        acq = pwsdt.AcqDir(folder)  # An object handling the contents of a single "Cell{X}" folder

        try:
            anls = acq.pws.loadAnalysis(analysisName)  # Load the analysis results from file.
        except:
            print(f"Analysis loading failed for {acq.filePath}")
            continue  # Skip to the next loop iteration

        roiSpecs = acq.getRois()  # A list of the names, numbers, and fileFormats of the ROIs in this acquisition

        for name, number, fformat in roiSpecs:  # Loop through every ROI.
            roi = acq.loadRoi(name, number, fformat)  # Load the ROI from file.
            plotHist(roi, anls.rms)  # Use the function defined above to plot a histogram




# -*- coding: utf-8 -*-
"""
Load the OPD from a previosly saved analysis result and plot it using a special multi-dimensional plotting widget.

@author: Nick Anthony
"""
if __name__ == '__main__':

    import pwspy.dataTypes as pwsdt
    from mpl_qt_viz.visualizers import PlotNd

    imageDirectory = r'\\backmanlabnas.myqnapcloud.com\home\Year3\ethanolTimeSeries\LTL20l_Pure\Track_3hrs\Cell3'

    acquisition = pwsdt.AcqDir(imageDirectory)

    roiSpecs = acquisition.getRois()
    print("ROIs:\n", roiSpecs)

    imCube = acquisition.pws.toDataClass()
    kCube = pwsdt.KCube.fromImCube(imCube)
    del imCube

    opd, opdValues = kCube.getOpd(isHannWindow = False, indexOpdStop = 50)

    ri = 1.37
    opdValues = opdValues / (2 * ri)

    plotWindow = PlotNd(opd, names=('y', 'x', 'depth'),
                        indices=(None, None, opdValues), title="Estimated Depth")


import sys
import os
# sys.path.append(os.path.join('..', 'src'))

from pwspy import analysis
import pwspy.dataTypes as pwsdt
from pwspy.utility.reflection import Material
import pytest

_analysisName = 'testAnalysis'
resourcesDir = os.path.join(os.path.split(__file__)[0], 'resources')
_acqDir = os.path.join(resourcesDir, 'seqTest', 'Cell1')
_refDir = os.path.join(resourcesDir, 'seqTest', 'Cell3')


def test_pws_analysis():
    settings = analysis.pws.PWSAnalysisSettings.loadDefaultSettings("Recommended")
    er = None  # TODO add other cases
    
    refAcq = pwsdt.AcqDir(_refDir)
    ref = refAcq.pws.toDataClass()  # TODO Test other cases of processing status (exposure normalized or not)
    anls = analysis.pws.PWSAnalysis(settings=settings, extraReflectance=er, ref=ref)
    
    acq = pwsdt.AcqDir(_acqDir)
    cube = acq.pws.toDataClass()  # TODO test various states of preprocessing
    results, warnings = anls.run(cube)

    with pytest.raises(OSError):
        acq.pws.saveAnalysis(results, _analysisName)  # The analysis already exists so an OSError should be thrown.
    acq.pws.saveAnalysis(results, _analysisName, overwrite=True)
    
    acq.pws.loadAnalysis(_analysisName)
    
    # TODO add assertions
    
def test_dynamics_analysis():
    er = None  # TODO add other cases

    settings = analysis.dynamics.DynamicsAnalysisSettings(
        cameraCorrection=None,
        extraReflectanceId=er.idTag if er is not None else None,
        referenceMaterial=Material.Water,
        numericalAperture=0.52,
        relativeUnits=True
    )

    refAcq = pwsdt.AcqDir(_refDir)
    ref = refAcq.dynamics.toDataClass()  # TODO Test other cases of processing status (exposure normalized or not)
    anls = analysis.dynamics.DynamicsAnalysis(settings=settings, extraReflectance=er, ref=ref)

    acq = pwsdt.AcqDir(_acqDir)
    cube = acq.dynamics.toDataClass()  # TODO test various states of preprocessing
    results, warnings = anls.run(cube)

    acq.dynamics.saveAnalysis(results, _analysisName)

    acq.dynamics.loadAnalysis(_analysisName)


def test_compilation():
    settings = analysis.compilation.GenericCompilerSettings(True)
    genComp = analysis.compilation.GenericRoiCompiler(settings)
    
    settings = analysis.compilation.PWSCompilerSettings(True, True, True)
    pwsComp = analysis.compilation.PWSRoiCompiler(settings)
    
    settings = analysis.compilation.DynamicsCompilerSettings(True, True, True)
    dynComp = analysis.compilation.DynamicsRoiCompiler(settings)
    
    acq = pwsdt.AcqDir(_acqDir)
    
    for roiSpecs in acq.getRois():
        roi = acq.loadRoi(*roiSpecs)
        results = (genComp.run(roi), 
                   pwsComp.run(acq.pws.loadAnalysis(_analysisName), roi),
                   dynComp.run(acq.dynamics.loadAnalysis(_analysisName), roi))
        print(results)


def test_sequence():
    from pwspy.utility.acquisition import loadDirectory
    seq, acqs = loadDirectory((os.path.join(resourcesDir, 'seqTest')))

if __name__ == '__main__':
    test_pws_analysis()
    a = 1